# 1 "src/dune_sexp/lexer.mll"
 
open! Stdune

open Stdune

module Token = struct
  type t =
    | Atom of Atom.t
    | Quoted_string of string
    | Lparen
    | Rparen
    | Eof
    | Template of Template.t
    | Comment of string list
end

type t = with_comments:bool -> Lexing.lexbuf -> Token.t

let error ?(delta = 0) lexbuf message =
  let start = Lexing.lexeme_start_p lexbuf in
  let loc =
    Loc.create ~start:{ start with pos_cnum = start.pos_cnum + delta }
      ~stop:(Lexing.lexeme_end_p lexbuf)
  in
  User_error.raise ~loc [ Pp.text message ]

let invalid_dune_or_jbuild lexbuf =
  let start = Lexing.lexeme_start_p lexbuf in
  let fname = Filename.basename start.pos_fname in
  error lexbuf (sprintf "Invalid %s file" fname)

type escape_sequence =
  | Newline
  | Other

let eval_decimal_char c = Char.code c - Char.code '0'

let eval_decimal_escape c1 c2 c3 =
  (eval_decimal_char c1 * 100)
  + (eval_decimal_char c2 * 10)
  + eval_decimal_char c3

let eval_hex_char c =
  match c with
  | '0' .. '9' -> Char.code c - Char.code '0'
  | 'a' .. 'f' -> Char.code c - Char.code 'a' + 10
  | 'A' .. 'F' -> Char.code c - Char.code 'A' + 10
  | _ -> -1

let eval_hex_escape c1 c2 = (eval_hex_char c1 * 16) + eval_hex_char c2

type block_string_line_kind =
  | With_escape_sequences
  | Raw

module Template = struct
  include Template

  let dummy_loc = Loc.create ~start:Lexing.dummy_pos ~stop:Lexing.dummy_pos

  let add_text parts s =
    match parts with
    | Template.Text s' :: parts -> Template.Text (s' ^ s) :: parts
    | _ -> Template.Text s :: parts

  let token parts ~quoted ~start (lexbuf : Lexing.lexbuf) =
    lexbuf.lex_start_p <- start;
    match parts with
    | [] | [Text ""] ->
      invalid_dune_or_jbuild lexbuf
    | [Text s] ->
      Token.Atom (Atom.of_string s)
    | _ ->
      Token.Template
        { quoted
        ; loc = dummy_loc
        ; parts = List.rev parts
        }

  module Buffer : sig
    val new_token : unit -> unit
    val get : unit -> Token.t
    val add_var : part -> unit
    val add_text : string -> unit
    val add_text_c : char -> unit
  end = struct
    type state =
      | String
      | Template of Template.part list

    let text_buf = Buffer.create 256

    let new_token () = Buffer.clear text_buf

    let take_buf () =
      let contents = Buffer.contents text_buf in
      Buffer.clear text_buf;
      contents

    let state = ref String

    let add_buf_to_parts parts =
      match take_buf () with
      | "" -> parts
      | t -> add_text parts t

    let get () =
      match !state with
      | String -> Token.Quoted_string (take_buf ())
      | Template parts ->
        state := String;
        begin match add_buf_to_parts parts with
        | [] -> assert false
        | [Text s] -> Quoted_string s
        | parts ->
          Token.Template
            { quoted = true
            ; loc = dummy_loc
            ; parts = List.rev parts
            }
        end

    let add_var v =
      match !state with
      | String ->
        state := Template (v :: add_buf_to_parts []);
      | Template parts ->
        let parts = add_buf_to_parts parts in
        state := Template (v::parts)

    let add_text   = Buffer.add_string text_buf
    let add_text_c = Buffer.add_char text_buf
  end
end

# 138 "src/dune_sexp/lexer.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base =
   "\000\000\249\255\250\255\251\255\252\255\001\000\006\000\255\255\
    \006\000\007\000\010\000\011\000\013\000\028\000\032\000\123\000\
    \254\255\065\000\188\000\254\255\255\255\014\000\252\255\253\255\
    \254\255\255\255\015\000\241\000\251\255\252\255\037\000\254\255\
    \255\255\018\000\253\255\021\000\071\000\243\000\191\000\254\255\
    \255\255\026\000\253\255\254\255\255\255\019\000\249\000\250\255\
    \251\255\252\255\021\000\041\000\254\255\255\255\253\255\250\000\
    \246\255\247\255\067\001\012\001\253\255\254\255\255\255\025\000\
    \045\001\252\255\090\001\249\255\028\000\209\001\253\255\254\255\
    \047\002\255\255\150\002\255\255";
  Lexing.lex_backtrk =
   "\007\000\255\255\255\255\255\255\255\255\002\000\001\000\255\255\
    \255\255\001\000\255\255\255\255\000\000\003\000\002\000\000\000\
    \255\255\002\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\003\000\255\255\255\255\255\255\003\000\255\255\
    \255\255\003\000\255\255\002\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\001\000\255\255\255\255\
    \255\255\255\255\004\000\004\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\007\000\005\000\255\255\255\255\255\255\008\000\
    \005\000\255\255\007\000\255\255\000\000\255\255\255\255\255\255\
    \002\000\255\255\002\000\255\255";
  Lexing.lex_default =
   "\255\255\000\000\000\000\000\000\000\000\005\000\255\255\000\000\
    \255\255\255\255\255\255\255\255\012\000\255\255\255\255\255\255\
    \000\000\255\255\255\255\000\000\000\000\022\000\000\000\000\000\
    \000\000\000\000\255\255\029\000\000\000\000\000\255\255\000\000\
    \000\000\255\255\000\000\255\255\255\255\255\255\255\255\000\000\
    \000\000\043\000\000\000\000\000\000\000\255\255\048\000\000\000\
    \000\000\000\000\255\255\255\255\000\000\000\000\000\000\057\000\
    \000\000\000\000\255\255\255\255\000\000\000\000\000\000\255\255\
    \255\255\000\000\255\255\000\000\255\255\070\000\000\000\000\000\
    \070\000\000\000\070\000\000\000";
  Lexing.lex_trans =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\006\000\007\000\255\255\006\000\008\000\255\255\006\000\
    \007\000\010\000\006\000\010\000\011\000\010\000\010\000\255\255\
    \025\000\025\000\255\255\026\000\032\000\044\000\037\000\049\000\
    \006\000\037\000\002\000\062\000\044\000\068\000\006\000\045\000\
    \004\000\003\000\010\000\000\000\000\000\000\000\024\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\037\000\000\000\036\000\
    \000\000\000\000\000\000\005\000\068\000\015\000\000\000\015\000\
    \015\000\014\000\015\000\015\000\000\000\012\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\000\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\016\000\015\000\018\000\015\000\015\000\
    \034\000\015\000\015\000\038\000\054\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\000\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\019\000\032\000\037\000\039\000\033\000\037\000\
    \001\000\255\255\000\000\049\000\062\000\000\000\050\000\063\000\
    \000\000\000\000\000\000\000\000\000\000\255\255\023\000\000\000\
    \000\000\000\000\000\000\037\000\000\000\036\000\030\000\000\000\
    \000\000\000\000\042\000\053\000\060\000\000\000\051\000\061\000\
    \000\000\060\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\000\000\000\000\000\000\000\000\
    \020\000\000\000\000\000\040\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\031\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\052\000\060\000\000\000\
    \000\000\000\000\000\000\000\000\060\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\000\000\
    \060\000\000\000\000\000\000\000\060\000\000\000\060\000\000\000\
    \000\000\000\000\058\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\000\000\000\000\000\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\028\000\072\000\000\000\072\000\072\000\000\000\072\000\
    \072\000\047\000\056\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\000\000\000\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\000\000\072\000\071\000\072\000\
    \072\000\000\000\072\000\072\000\000\000\072\000\072\000\000\000\
    \000\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\074\000\000\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\000\000\072\000\073\000\072\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\074\000\
    \000\000\074\000\074\000\000\000\074\000\074\000\000\000\000\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\071\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\000\000\074\000\075\000\074\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255";
  Lexing.lex_check =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\005\000\000\000\000\000\005\000\006\000\
    \008\000\009\000\006\000\010\000\009\000\011\000\010\000\012\000\
    \021\000\026\000\012\000\021\000\033\000\045\000\035\000\050\000\
    \000\000\035\000\000\000\063\000\041\000\068\000\006\000\041\000\
    \000\000\000\000\010\000\255\255\255\255\255\255\021\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\035\000\255\255\035\000\
    \255\255\255\255\255\255\000\000\068\000\013\000\255\255\013\000\
    \013\000\013\000\013\000\013\000\255\255\010\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\255\255\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\014\000\015\000\017\000\015\000\015\000\
    \030\000\015\000\015\000\036\000\051\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\255\255\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\018\000\027\000\037\000\038\000\027\000\037\000\
    \000\000\005\000\255\255\046\000\055\000\255\255\046\000\055\000\
    \255\255\255\255\255\255\255\255\255\255\012\000\021\000\255\255\
    \255\255\255\255\255\255\037\000\255\255\037\000\027\000\255\255\
    \255\255\255\255\041\000\046\000\055\000\255\255\046\000\055\000\
    \255\255\055\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\255\255\255\255\255\255\255\255\
    \018\000\255\255\255\255\038\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\027\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\046\000\055\000\255\255\
    \255\255\255\255\255\255\255\255\055\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\255\255\
    \055\000\255\255\255\255\255\255\055\000\255\255\055\000\255\255\
    \255\255\255\255\055\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\058\000\058\000\058\000\058\000\
    \058\000\058\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\066\000\066\000\066\000\066\000\066\000\
    \066\000\255\255\255\255\255\255\058\000\058\000\058\000\058\000\
    \058\000\058\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\066\000\066\000\066\000\066\000\066\000\
    \066\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\027\000\069\000\255\255\069\000\069\000\255\255\069\000\
    \069\000\046\000\055\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\255\255\255\255\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\255\255\069\000\069\000\069\000\
    \072\000\255\255\072\000\072\000\255\255\072\000\072\000\255\255\
    \255\255\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\255\255\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\255\255\072\000\072\000\072\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\074\000\
    \255\255\074\000\074\000\255\255\074\000\074\000\255\255\255\255\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\069\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\255\255\074\000\074\000\074\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\072\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\074\000";
  Lexing.lex_base_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\014\000\000\000\
    \092\000\046\000\186\000\069\000";
  Lexing.lex_backtrk_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\004\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \014\000\000\000\051\000\000\000";
  Lexing.lex_default_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \021\000\000\000\000\000\000\000";
  Lexing.lex_trans_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\007\000\000\000\007\000\007\000\000\000\007\000\007\000\
    \000\000\000\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\000\000\001\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\000\000\007\000\026\000\007\000\026\000\
    \026\000\000\000\026\000\026\000\000\000\000\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\037\000\000\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\000\000\
    \026\000\000\000\026\000\058\000\000\000\058\000\058\000\000\000\
    \058\000\058\000\000\000\000\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\000\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\000\000\058\000\000\000\
    \058\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000";
  Lexing.lex_check_code =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\069\000\255\255\069\000\069\000\255\255\069\000\069\000\
    \255\255\255\255\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\255\255\010\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\255\255\069\000\072\000\069\000\072\000\
    \072\000\255\255\072\000\072\000\255\255\255\255\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\255\255\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\255\255\
    \072\000\072\000\072\000\074\000\255\255\074\000\074\000\255\255\
    \074\000\074\000\255\255\255\255\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\255\255\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\255\255\074\000\255\255\
    \074\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\072\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255";
  Lexing.lex_code =
   "\255\001\255\255\000\001\255\008\255\007\255\006\255\255\002\003\
    \001\005\000\004\255\005\008\004\007\255\005\008\004\007\008\255\
    \007\255\006\255\255\010\255\009\255\005\255\004\255\255\002\255\
    \000\006\255\002\003\001\008\000\007\255\008\005\007\004\005\255\
    \010\255\004\255\255\002\009\001\010\000\006\255";
}

let rec token with_comments lexbuf =
   __ocaml_lex_token_rec with_comments lexbuf 0
and __ocaml_lex_token_rec with_comments lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 149 "src/dune_sexp/lexer.mll"
    ( Lexing.new_line lexbuf; token with_comments lexbuf )
# 567 "src/dune_sexp/lexer.ml"

  | 1 ->
# 151 "src/dune_sexp/lexer.mll"
    ( token with_comments lexbuf )
# 572 "src/dune_sexp/lexer.ml"

  | 2 ->
# 153 "src/dune_sexp/lexer.mll"
    ( if with_comments then
        comment_trail [String.drop (Lexing.lexeme lexbuf) 1] lexbuf
      else
        token with_comments lexbuf
    )
# 581 "src/dune_sexp/lexer.ml"

  | 3 ->
# 159 "src/dune_sexp/lexer.mll"
    ( Token.Lparen )
# 586 "src/dune_sexp/lexer.ml"

  | 4 ->
# 161 "src/dune_sexp/lexer.mll"
    ( Rparen )
# 591 "src/dune_sexp/lexer.ml"

  | 5 ->
# 163 "src/dune_sexp/lexer.mll"
    ( let start = Lexing.lexeme_start_p lexbuf in
      Template.Buffer.new_token ();
      let token = start_quoted_string lexbuf in
      lexbuf.lex_start_p <- start;
      token
    )
# 601 "src/dune_sexp/lexer.ml"

  | 6 ->
# 170 "src/dune_sexp/lexer.mll"
    ( Eof )
# 606 "src/dune_sexp/lexer.ml"

  | 7 ->
# 172 "src/dune_sexp/lexer.mll"
    ( atom [] (Lexing.lexeme_start_p lexbuf) lexbuf )
# 611 "src/dune_sexp/lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_token_rec with_comments lexbuf __ocaml_lex_state

and comment_trail acc lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 2 (-1); __ocaml_lex_comment_trail_rec acc lexbuf 9
and __ocaml_lex_comment_trail_rec acc lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 175 "src/dune_sexp/lexer.mll"
                                        s
# 624 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 176 "src/dune_sexp/lexer.mll"
    ( comment_trail (s :: acc) lexbuf )
# 628 "src/dune_sexp/lexer.ml"

  | 1 ->
# 178 "src/dune_sexp/lexer.mll"
    ( Token.Comment (List.rev acc) )
# 633 "src/dune_sexp/lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_comment_trail_rec acc lexbuf __ocaml_lex_state

and atom acc start lexbuf =
   __ocaml_lex_atom_rec acc start lexbuf 13
and __ocaml_lex_atom_rec acc start lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 181 "src/dune_sexp/lexer.mll"
                          s
# 646 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 182 "src/dune_sexp/lexer.mll"
    ( atom (Template.add_text acc s) start lexbuf )
# 650 "src/dune_sexp/lexer.ml"

  | 1 ->
# 184 "src/dune_sexp/lexer.mll"
    ( atom ((template_variable lexbuf) :: acc) start lexbuf )
# 655 "src/dune_sexp/lexer.ml"

  | 2 ->
# 186 "src/dune_sexp/lexer.mll"
    ( atom (Template.add_text acc "%") start lexbuf )
# 660 "src/dune_sexp/lexer.ml"

  | 3 ->
# 188 "src/dune_sexp/lexer.mll"
    ( Template.token acc ~quoted:false ~start lexbuf )
# 665 "src/dune_sexp/lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_atom_rec acc start lexbuf __ocaml_lex_state

and start_quoted_string lexbuf =
   __ocaml_lex_start_quoted_string_rec lexbuf 17
and __ocaml_lex_start_quoted_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 192 "src/dune_sexp/lexer.mll"
    ( block_string_start With_escape_sequences lexbuf )
# 677 "src/dune_sexp/lexer.ml"

  | 1 ->
# 194 "src/dune_sexp/lexer.mll"
    ( block_string_start Raw lexbuf )
# 682 "src/dune_sexp/lexer.ml"

  | 2 ->
# 196 "src/dune_sexp/lexer.mll"
    ( quoted_string lexbuf )
# 687 "src/dune_sexp/lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_start_quoted_string_rec lexbuf __ocaml_lex_state

and block_string_start kind lexbuf =
   __ocaml_lex_block_string_start_rec kind lexbuf 21
and __ocaml_lex_block_string_start_rec kind lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 199 "src/dune_sexp/lexer.mll"
               s
# 700 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 200 "src/dune_sexp/lexer.mll"
    ( Lexing.new_line lexbuf;
      Template.Buffer.add_text s;
      block_string_after_newline lexbuf
    )
# 707 "src/dune_sexp/lexer.ml"

  | 1 ->
# 205 "src/dune_sexp/lexer.mll"
    ( match kind with
      | With_escape_sequences -> block_string lexbuf
      | Raw -> raw_block_string lexbuf
    )
# 715 "src/dune_sexp/lexer.ml"

  | 2 ->
# 210 "src/dune_sexp/lexer.mll"
    ( Template.Buffer.get () )
# 720 "src/dune_sexp/lexer.ml"

  | 3 ->
# 212 "src/dune_sexp/lexer.mll"
    ( error lexbuf "There must be at least one space after \"\\|"
    )
# 726 "src/dune_sexp/lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_block_string_start_rec kind lexbuf __ocaml_lex_state

and block_string lexbuf =
   __ocaml_lex_block_string_rec lexbuf 27
and __ocaml_lex_block_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 216 "src/dune_sexp/lexer.mll"
               s
# 739 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 217 "src/dune_sexp/lexer.mll"
    ( Lexing.new_line lexbuf;
      Template.Buffer.add_text s;
      block_string_after_newline lexbuf
    )
# 746 "src/dune_sexp/lexer.ml"

  | 1 ->
# 222 "src/dune_sexp/lexer.mll"
    ( match escape_sequence lexbuf with
      | Newline -> block_string_after_newline lexbuf
      | Other   -> block_string               lexbuf
    )
# 754 "src/dune_sexp/lexer.ml"

  | 2 ->
# 226 "src/dune_sexp/lexer.mll"
         (
      let var = template_variable lexbuf in
      Template.Buffer.add_var var;
      block_string lexbuf
    )
# 763 "src/dune_sexp/lexer.ml"

  | 3 ->
let
# 231 "src/dune_sexp/lexer.mll"
         c
# 769 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 232 "src/dune_sexp/lexer.mll"
    ( Template.Buffer.add_text_c c;
      block_string lexbuf
    )
# 775 "src/dune_sexp/lexer.ml"

  | 4 ->
# 236 "src/dune_sexp/lexer.mll"
    ( Template.Buffer.get ()
    )
# 781 "src/dune_sexp/lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_block_string_rec lexbuf __ocaml_lex_state

and block_string_after_newline lexbuf =
   __ocaml_lex_block_string_after_newline_rec lexbuf 35
and __ocaml_lex_block_string_after_newline_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 241 "src/dune_sexp/lexer.mll"
    ( block_string_start With_escape_sequences lexbuf )
# 793 "src/dune_sexp/lexer.ml"

  | 1 ->
# 243 "src/dune_sexp/lexer.mll"
    ( block_string_start Raw lexbuf )
# 798 "src/dune_sexp/lexer.ml"

  | 2 ->
# 245 "src/dune_sexp/lexer.mll"
    ( Template.Buffer.get ()
    )
# 804 "src/dune_sexp/lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_block_string_after_newline_rec lexbuf __ocaml_lex_state

and raw_block_string lexbuf =
   __ocaml_lex_raw_block_string_rec lexbuf 41
and __ocaml_lex_raw_block_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 249 "src/dune_sexp/lexer.mll"
               s
# 817 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 250 "src/dune_sexp/lexer.mll"
    ( Lexing.new_line lexbuf;
      Template.Buffer.add_text s;
      block_string_after_newline lexbuf
    )
# 824 "src/dune_sexp/lexer.ml"

  | 1 ->
let
# 254 "src/dune_sexp/lexer.mll"
         c
# 830 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 255 "src/dune_sexp/lexer.mll"
    ( Template.Buffer.add_text_c c;
      raw_block_string lexbuf
    )
# 836 "src/dune_sexp/lexer.ml"

  | 2 ->
# 259 "src/dune_sexp/lexer.mll"
    ( Template.Buffer.get ()
    )
# 842 "src/dune_sexp/lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_raw_block_string_rec lexbuf __ocaml_lex_state

and quoted_string lexbuf =
   __ocaml_lex_quoted_string_rec lexbuf 46
and __ocaml_lex_quoted_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 264 "src/dune_sexp/lexer.mll"
    ( Template.Buffer.get () )
# 854 "src/dune_sexp/lexer.ml"

  | 1 ->
# 266 "src/dune_sexp/lexer.mll"
    ( match escape_sequence lexbuf with
      | Newline -> quoted_string_after_escaped_newline lexbuf
      | Other   -> quoted_string                       lexbuf
    )
# 862 "src/dune_sexp/lexer.ml"

  | 2 ->
# 271 "src/dune_sexp/lexer.mll"
    ( Template.Buffer.add_var (template_variable lexbuf);
      quoted_string lexbuf
    )
# 869 "src/dune_sexp/lexer.ml"

  | 3 ->
let
# 274 "src/dune_sexp/lexer.mll"
               s
# 875 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 275 "src/dune_sexp/lexer.mll"
    ( Lexing.new_line lexbuf;
      Template.Buffer.add_text s;
      quoted_string lexbuf
    )
# 882 "src/dune_sexp/lexer.ml"

  | 4 ->
let
# 279 "src/dune_sexp/lexer.mll"
         c
# 888 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 280 "src/dune_sexp/lexer.mll"
    ( Template.Buffer.add_text_c c;
      quoted_string lexbuf
    )
# 894 "src/dune_sexp/lexer.ml"

  | 5 ->
# 284 "src/dune_sexp/lexer.mll"
    ( error lexbuf "unterminated quoted string"
    )
# 900 "src/dune_sexp/lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_quoted_string_rec lexbuf __ocaml_lex_state

and escape_sequence lexbuf =
   __ocaml_lex_escape_sequence_rec lexbuf 55
and __ocaml_lex_escape_sequence_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 289 "src/dune_sexp/lexer.mll"
    ( Lexing.new_line lexbuf;
      Newline )
# 913 "src/dune_sexp/lexer.ml"

  | 1 ->
# 292 "src/dune_sexp/lexer.mll"
    ( Template.Buffer.add_text_c '%';
      Other
    )
# 920 "src/dune_sexp/lexer.ml"

  | 2 ->
let
# 295 "src/dune_sexp/lexer.mll"
                                       c
# 926 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 296 "src/dune_sexp/lexer.mll"
    ( let c =
        match c with
        | 'n' -> '\n'
        | 'r' -> '\r'
        | 'b' -> '\b'
        | 't' -> '\t'
        | _   -> c
      in
      Template.Buffer.add_text_c c;
      Other
    )
# 940 "src/dune_sexp/lexer.ml"

  | 3 ->
let
# 307 "src/dune_sexp/lexer.mll"
              c1
# 946 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos
and
# 307 "src/dune_sexp/lexer.mll"
                            c2
# 951 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1)
and
# 307 "src/dune_sexp/lexer.mll"
                                          c3
# 956 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 308 "src/dune_sexp/lexer.mll"
    ( let v = eval_decimal_escape c1 c2 c3 in
      if v > 255 then
        error lexbuf "escape sequence in quoted string out of range"
          ~delta:(-1);
      Template.Buffer.add_text_c (Char.chr v);
      Other
    )
# 966 "src/dune_sexp/lexer.ml"

  | 4 ->
# 316 "src/dune_sexp/lexer.mll"
    ( error lexbuf "escape sequence in quoted string out of range" ~delta:(-1);
    )
# 972 "src/dune_sexp/lexer.ml"

  | 5 ->
# 319 "src/dune_sexp/lexer.mll"
    ( error lexbuf "unterminated decimal escape sequence" ~delta:(-1);
    )
# 978 "src/dune_sexp/lexer.ml"

  | 6 ->
let
# 321 "src/dune_sexp/lexer.mll"
                     c1
# 984 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1)
and
# 321 "src/dune_sexp/lexer.mll"
                                      c2
# 989 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 322 "src/dune_sexp/lexer.mll"
    ( let v = eval_hex_escape c1 c2 in
      Template.Buffer.add_text_c (Char.chr v);
      Other
    )
# 996 "src/dune_sexp/lexer.ml"

  | 7 ->
# 327 "src/dune_sexp/lexer.mll"
    ( error lexbuf "unterminated hexadecimal escape sequence" ~delta:(-1);
    )
# 1002 "src/dune_sexp/lexer.ml"

  | 8 ->
# 330 "src/dune_sexp/lexer.mll"
    ( error lexbuf "unknown escape sequence" ~delta:(-1);
    )
# 1008 "src/dune_sexp/lexer.ml"

  | 9 ->
# 333 "src/dune_sexp/lexer.mll"
    ( error lexbuf "unterminated escape sequence" ~delta:(-1);
    )
# 1014 "src/dune_sexp/lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_escape_sequence_rec lexbuf __ocaml_lex_state

and quoted_string_after_escaped_newline lexbuf =
   __ocaml_lex_quoted_string_after_escaped_newline_rec lexbuf 68
and __ocaml_lex_quoted_string_after_escaped_newline_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 338 "src/dune_sexp/lexer.mll"
    ( quoted_string lexbuf )
# 1026 "src/dune_sexp/lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_quoted_string_after_escaped_newline_rec lexbuf __ocaml_lex_state

and template_variable lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 11 (-1);(* L=3 [5] <- p ; [4] <- p ; [3] <- p ;  *)
  lexbuf.Lexing.lex_mem.(5) <- lexbuf.Lexing.lex_curr_pos ;
  lexbuf.Lexing.lex_mem.(4) <- lexbuf.Lexing.lex_curr_pos ;
  lexbuf.Lexing.lex_mem.(3) <- lexbuf.Lexing.lex_curr_pos ;
 __ocaml_lex_template_variable_rec lexbuf 69
and __ocaml_lex_template_variable_rec lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 341 "src/dune_sexp/lexer.mll"
                      name
# 1043 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_mem.(0)
and
# 341 "src/dune_sexp/lexer.mll"
                                                           payload
# 1048 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme_opt lexbuf lexbuf.Lexing.lex_mem.(2) lexbuf.Lexing.lex_mem.(1) in
# 342 "src/dune_sexp/lexer.mll"
    ( let payload =
        match payload with
        | Some "" -> error lexbuf "payload after : in variable cannot be empty"
        | p -> p
      in
      let start = Lexing.lexeme_start_p lexbuf in
      (* -2 to account for the "%{" *)
      let start = { start with pos_cnum = start.pos_cnum - 2 } in
      let loc = Loc.create ~start ~stop:(Lexing.lexeme_end_p lexbuf) in
      Template.Pform { loc ; name ; payload = Option.map ~f:Template.Pform.Payload.of_string payload }
  )
# 1062 "src/dune_sexp/lexer.ml"

  | 1 ->
# 354 "src/dune_sexp/lexer.mll"
    ( error lexbuf "%{...} forms cannot be empty" )
# 1067 "src/dune_sexp/lexer.ml"

  | 2 ->
let
# 355 "src/dune_sexp/lexer.mll"
                      skip
# 1073 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(2) lexbuf.Lexing.lex_mem.(1)
and
# 355 "src/dune_sexp/lexer.mll"
                                  other
# 1078 "src/dune_sexp/lexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_mem.(0) in
# 357 "src/dune_sexp/lexer.mll"
  (
    error
      ~delta:(String.length skip)
      lexbuf
      (Printf.sprintf "The character %C is not allowed inside %%{...} forms" other)
  )
# 1087 "src/dune_sexp/lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_template_variable_rec lexbuf __ocaml_lex_state

;;

# 364 "src/dune_sexp/lexer.mll"
 
  let token ~with_comments lexbuf = token with_comments lexbuf

# 1098 "src/dune_sexp/lexer.ml"
